/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
						ip_rcv(softirq)				ip_output()
							|							|
						NF_HOOK()						|
							|							|
						NF_HOOK_THRESH()			NF_HOOK_COND()
							|							|
							|---------------------------|
							|
 okfn() <-- ret == 1 -- nf_hook_thresh() <---------------------------------|
							|											   |
						nf_hook_slow()									   |
							|											   |
						nf_iterate(&nf_hooks[pf][hook]) <-------------|    |
							|										  |    |
							|-- NF_QUEUE --> nf_queue() -- ret == 0 --|	   |
							|					|						   |
							|					|-- -------ret != 0 -------|
							|											   |
							|-- NF_ACCEPT/NF_STOP -------- ret == 1 -------|
							|											   |
							|-- NF_DROP ------------------ ret == -EPERM --|
*/
#include "hdrs.h"
#include "pkt.h"
#include "hooks.h"
#include "nf_defrag_ipv6.h"

/*
 *	Globals
 */
struct ipf_queue g_main_queue;
struct ipf_queue g_pending_queue;

/*
 *	Locals
 */
static int nf_hooks_registered = 0;
static int nf_queue_handler_registered = 0;
static int nf_hook1_bytes_to_replace = 0;
static int nf_hook2_bytes_to_replace = 0;
static int snort_ready = 0;
static const struct nf_queue_handler *queue_handler[1];
static DEFINE_MUTEX(queue_handler_mutex);

static void nat_decode_session(struct sk_buff *skb, struct flowi *fl)
{
	struct flowi4 *fl4 = &fl->u.ip4;

	if ( ! skb->kct.used) {
		return;
	}

	fl4->daddr = skb->kct.daddr;

	TRACE_DBG("%s skb %p flowi: %d.%d.%d.%d > %d.%d.%d.%d\n", __func__, skb, IP_ADDRESS_STRING(fl4->saddr), IP_ADDRESS_STRING(fl4->daddr));
}

static u_int8_t
ip_filter_check_timeout(
	void
	)
{
	unsigned long t = get_seconds();

	if ((g_linipf_cfg.start_time + START_TIMEOUT) < t) {
		g_linipf_cfg.timeout_expired = 1;
		TRACE(LOG_INFO, "%s: timeout expired at %u", __func__, t);
		return 1;
	}

	return 0;
}

static u_int32_t
ip_filter_pre_process_packet(
	struct sk_buff *skb
	)
{
	if (!g_linipf_cfg.block_traffic_when_api_is_closed) {
		TRACE_DBG("%s: skb %p accepted due to internal policy.\n", __func__, skb);
		return NF_ACCEPT;
	}

	if (g_linipf_cfg.timeout_expired) {
		TRACE_DBG("%s: skb %p accepted due to expired timeout.\n", __func__, skb);
		return NF_ACCEPT;
	}

	if (ip_filter_check_timeout()) {
		TRACE_DBG("%s: skb %p accepted due to expired timeout.\n", __func__, skb);
		return NF_ACCEPT;
	}

	TRACE_DBG("%s: skb %p dropped.\n", __func__, skb);

	return NF_DROP;
}

static unsigned int ip_hook(const struct nf_hook_ops *ops, struct sk_buff *skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	const struct net_device *dev = ((ops && ops->hooknum == NF_IP_PRE_ROUTING) ? (in) : (out));

	if ( ! ops) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to null ops.", __func__, skb);
		return NF_DROP;
	}

	TRACE_DBG("%s: skb %p, len %d, proto 0x%04x, %s on '%s'", __func__, skb, skb->len, htons(skb->protocol), ((ops->hooknum == NF_IP_PRE_ROUTING) ? ("received") : ("sent")), ((dev) ? (dev->name) : ("???\n")));

	if (skb->protocol != htons(ETH_P_IPV6) && skb->protocol != htons(ETH_P_IP)) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to incorrect protocol (0x%04x).", __func__, skb, htons(skb->protocol));
		return NF_DROP;
	}

	dump_ip_packet("ip_hook()", skb);

	if (!dev) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to null net device (in %p, out %p).", __func__, skb, in, out);
		return NF_DROP;
	}
	/* packet was indicated via loopback device ? */
	if (dev->flags & IFF_LOOPBACK) {
		TRACE_DBG("%s: skb %p accepted due to indication on loopback device.\n", __func__, skb);
		return NF_ACCEPT;
	}
	/* is our char device opened ? */
	if (!g_device_opened) {
		return ip_filter_pre_process_packet(skb);
	}
	/* Check if policy routing was applied on packet or packet was sent by pkt_tx_ip() */
	/* Check if packet was previously indicated by pkt_tx_rx_eth() ? */
	if (skb->mark == PKT_SOCKET_TAG) {
		TRACE_DBG("%s: skb %p accepted due to policy routing or indication via our function.\n", __func__, skb);
		skb->mark = 0;
		return NF_ACCEPT;
	}

	/* gather fragments (if any) */
	{
		struct sk_buff *reasm = ip_gather_frags(skb);
		if (!reasm) {
			return NF_STOLEN;
		}
		/* it will happen only on IPv6 packets */
		if (reasm != skb) {
			int ret = 0;
			
			nf_ct_frag6_consume_orig(reasm);	
			
			if ((ret = NF_HOOK_THRESH(NFPROTO_IPV6, ops->hooknum, reasm, (struct net_device *)in, (struct net_device *)out, okfn, NF_IP6_PRI_CONNTRACK_DEFRAG + 1))) {
				TRACE(LOG_WARN, "%s: fragment: %d, skb %p.", __func__, ret, reasm);
			}
			
			return NF_STOLEN;
		}
	}

	/*
	 * Possible return values:
	 *
	 * NF_ACCEPT: continue traversal as normal.
	 * NF_DROP: drop the packet; don't continue traversal.
	 * NF_STOLEN: I've taken over the packet; don't continue traversal.
	 * NF_QUEUE: queue the packet (usually for userspace handling).
	 * NF_REPEAT: call this hook again.
	 */
	//TRACE_DBG("%s: skb %p, len %d inserted into netfilter queue.\n", __func__, skb, skb->len);
	return NF_QUEUE;
}

static unsigned int ipv4_hook(const struct nf_hook_ops *ops, struct sk_buff *skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	if (skb->len < sizeof(struct iphdr)) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to incorrect len (%d).", __func__, skb, skb->len);
		return NF_DROP;
	}
	if (skb->protocol != htons(ETH_P_IP)) {
		skb->protocol = htons(ETH_P_IP);
	}
	return ip_hook(ops, skb, in, out, okfn);
}

static unsigned int ipv6_hook(const struct nf_hook_ops *ops, struct sk_buff *skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	if (skb->len < sizeof(struct ipv6hdr)) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to incorrect len (%d).", __func__, skb, skb->len);
		return NF_DROP;
	}
	/* on output path, protocol is filled in ip6_finish_output2 */
	if (skb->protocol != htons(ETH_P_IPV6)) {
		skb->protocol = htons(ETH_P_IPV6);
	}
	if (skb->kdata.type == NETWORK_PACKET && skb->kdata.reasm) {
		TRACE_DBG("%s: skb %p inserted into netfilter queue.\n", __func__, skb);
		return NF_QUEUE;
	}
	return ip_hook(ops, skb, in, out, okfn);
}

static unsigned int ip_snort_hook(const struct nf_hook_ops *ops, struct sk_buff *skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	const struct net_device *dev = ((ops && ops->hooknum == NF_IP_PRE_ROUTING) ? (in) : (out));
	const struct nf_queue_handler *qh = NULL;
	bool queue = false;

	if ( ! ops) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to null ops.", __func__, skb);
		return NF_DROP;
	}

	TRACE_DBG("%s: skb %p, len %d, %s on '%s'", __func__, skb, skb->len, ((ops->hooknum == NF_IP_PRE_ROUTING) ? ("received") : ("sent")), ((dev) ? (dev->name) : ("???\n")));

	dump_ip_packet("ip_snort_hook()", skb);

	if (!dev) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to null net device (in %p, out %p).", __func__, skb, in, out);
		return NF_DROP;
	}

	/* Check if packet was previously indicated by pkt_tx_rx_eth() ? */
	if (ops->hooknum == NF_IP_PRE_ROUTING && skb->mark == PKT_SOCKET_TAG) {
		TRACE_DBG("%s: skb %p accepted due to indication via our function.\n", __func__, skb);
		/* Do not clear mark flag otherwise ip_hook will not recognize this packet as sent from pkt_tx_rx_eth */
		//skb->mark = 0;
		return NF_ACCEPT;
	}

	rcu_read_lock();

	qh = rcu_dereference(queue_handler[0]);

	queue = ((qh && snort_ready) ? (true) : (false));

	rcu_read_unlock();

	if (queue) {
		struct net_device_extension *ext = netdev_get_extension(dev);
		if (!ext) {
			TRACE(LOG_WARN, "%s: skb %p dropped due to missing device extension on '%s'.", __func__, skb, dev->name);
			return NF_DROP;
		}
		if (ext->mode & NF_IFACE_MODE_ENABLED) {
			TRACE_DBG("%s: skb %p inserted into kernel netfilter queue.\n", __func__, skb);
			return NF_QUEUE;
		}
	}

	TRACE_DBG("%s: skb %p accepted.\n", __func__, skb);
	return NF_ACCEPT;
}

static int __enqueue_packet(struct nf_queue_entry *nf_data, unsigned int queuenum)
{
	struct sk_buff *skb = nf_data->skb;
	unsigned int hooknum = nf_data->hook;
	bool snort_path = (nf_data->elem->hook == ip_snort_hook);

	/* ipv4_hook()/ipv6_hook() returns NF_QUEUE? */
	if (!snort_path) {
		if (skb->kdata.type == NETWORK_PACKET) {
			/* skb->nfct_reasm can't be valid on snort path - ipv6 packet is reassembled only in ip_hook() */
			if (skb->protocol == htons(ETH_P_IPV6) && skb->kdata.reasm) {
				struct sk_buff *reasm = skb->kdata.reasm;

				TRACE_DBG("%s: skb %p has reasm packet %p, next %p\n", __func__, skb, reasm, skb->kdata.next);

				/* other members filled in nf_ct_frag6_output() */
				skb->kdata.nf_data = nf_data;

				/* wait for last fragment */
				if (skb->kdata.next) {
					return 0;
				}
				/* exchange reasm packet with last indicated fragment */
				reasm->kdata.nf_data = nf_data;
				skb = nf_data->skb = reasm;
			} else {
				TRACE(LOG_ERR, "%s:  ### skb %p has our tag?!? how ?!? ###", __func__, skb);
				/* dump stack, only in debug version */
				dump_stack();
				nf_reinject(nf_data, NF_DROP);
				return 0;
			}
		} else {
			init_kipf_skb_data(skb, NETWORK_PACKET);
			skb->kdata.nf_data = nf_data;
			skb->kdata.received = ((hooknum == NF_IP_PRE_ROUTING) ? (true) :(false));
		}
	}

	/*
	 * Invalidate hardware checksum when packet is to be mangled, and complete checksum manually on outgoing path.
	 * moved due to [Bug 58820] - investigate suspicious kernel warning
	 */
	if ( ! skb_is_gso(skb) && (skb->ip_summed == CHECKSUM_PARTIAL || skb->ip_summed == CHECKSUM_COMPLETE)) {
		int err = 0;
		if ((err = skb_checksum_help(skb))) {
			TRACE(LOG_WARN, "%s: skb %p dropped due to error in checksum invalidation. ip_summed %d, err %d", __func__, skb, skb->ip_summed, err);
			goto drop;
		}
	}

	/* ip_snort_hook() returns NF_QUEUE? */
	if (snort_path) {
		const struct nf_queue_handler *qh = NULL;
		/* rcu_read_lock() is already called by caller (func __nf_queue in \net\netfilter\nf_queue.c) */
		if (!(qh = rcu_dereference(queue_handler[0]))) {
			TRACE(LOG_WARN, "%s: rcu_dereference() failed, skb %p.", __func__, skb);
			goto drop;
		}
		TRACE_DBG("%s: calling original enqueue packet handler on skb %p.\n", __func__, skb);

		return (qh->outfn(nf_data, queuenum));
		/* rcu_read_unlock() will be called immediately by caller */
	}

	/* limit for max_buffers_in_queue reached? */
	if (queue_len(&g_main_queue) + 1 > g_linipf_cfg.max_buffers_in_queue) {
		TRACE_DBG("%s: skb %p dropped due to limit of queued buffers in main queue (%d).\n", __func__, skb, g_linipf_cfg.max_buffers_in_queue);
		goto drop;
	}
	/* limit for max_buffers reached? */
	if (queue_len(&g_main_queue) + queue_len(&g_pending_queue) + 1 > g_linipf_cfg.max_buffers) {
		TRACE_DBG("%s: skb %p dropped due to limit of queued buffers (%d).\n", __func__, skb, g_linipf_cfg.max_buffers);
		goto drop;
	}

	TRACE_DBG("%s: skb %p, len %d, dev '%s', head_len %d, frag_list %p, ip_summed %d, mac_len %d, head %p, mac_hdr %p, net_hdr %p\n",
		__func__, skb, skb->len, ((skb->dev) ? (skb->dev->name) : ("???")), skb_headlen(skb), skb_shinfo(skb)->frag_list,
		skb->ip_summed, skb->mac_len, skb->head, skb_mac_header(skb), skb_network_header(skb));

	/* insert into main packet queue */
	if (queue_insert(&g_main_queue, skb, 1)) {
		TRACE(LOG_WARN, "%s: skb %p dropped due to disabled main queue.", __func__, skb);
		goto drop;
	}

	//TRACE_DBG("%s: skb %p inserted into main queue.\n", __func__, skb);
	io_dev_wakeup();
	return 0;
drop:
	/* on snort path skb->cb[] isn't initialized properly */
	if (snort_path) {
		nf_reinject(nf_data, NF_DROP);
	} else {
		pkt_verdict(skb, NF_DROP);
	}
	return 0;
}

/* this is borrowed by borrowes from net/netfilter/nfnetlink_queue_core.c */
/* we do not have CONFIG_NF_BRIDGE defined, but we also need the mac_header */
static void nf_bridge_adjust_skb_data(struct sk_buff *skb)
{
	if (skb_mac_header_was_set(skb)) {
		__skb_push(skb, skb->network_header - skb->mac_header);
	}
}

static void nf_bridge_adjust_segmented_data(struct sk_buff *skb)
{
	if (skb_mac_header_was_set(skb)) {
		__skb_pull(skb, skb->network_header - skb->mac_header);
	}
}

static void free_entry(struct nf_queue_entry *entry)
{
	nf_queue_entry_release_refs(entry);
	kfree(entry);
}

static struct nf_queue_entry *nf_queue_entry_dup(struct nf_queue_entry *e)
{
	struct nf_queue_entry *entry = kmemdup(e, e->size, GFP_ATOMIC);
	if (entry) {
		if (nf_queue_entry_get_refs(entry)) {
			return entry;
		}
		kfree(entry);
	}
	return NULL;
}

static int __enqueue_packet_gso(struct sk_buff *skb, struct nf_queue_entry *entry, unsigned int queuenum)
{
	int ret = -ENOMEM;
	struct nf_queue_entry *entry_seg;

	nf_bridge_adjust_segmented_data(skb);

	if (skb->next == NULL) { /* last packet, no need to copy entry */
		struct sk_buff *gso_skb = entry->skb;
		entry->skb = skb;
		ret = __enqueue_packet(entry, queuenum);
		if (ret) {
			entry->skb = gso_skb;
		}
		return ret;
	}

	skb->next = NULL;
	entry_seg = nf_queue_entry_dup(entry);

	if (entry_seg) {
		entry_seg->skb = skb;
		ret = __enqueue_packet(entry_seg, queuenum);
		if (ret) {
			free_entry(entry_seg);
		}
	}
	return ret;
}

static int enqueue_packet(struct nf_queue_entry *entry, unsigned int queuenum)
{
	unsigned int queued;
	struct sk_buff *skb, *segs;
	int err = -ENOBUFS;

	if ( ! skb_is_gso(entry->skb) || g_linipf_cfg.gso_enabled) {
		return __enqueue_packet(entry, queuenum);
	}

	skb = entry->skb;

	switch (entry->pf) {
		case NFPROTO_IPV4:
			skb->protocol = htons(ETH_P_IP);
			break;

		case NFPROTO_IPV6:
			skb->protocol = htons(ETH_P_IPV6);
			break;
	}

	nf_bridge_adjust_skb_data(skb);
	segs = skb_gso_segment(skb, 0);

	if (IS_ERR(segs)) {
		goto out_err;
	}

	queued = 0;
	err = 0;
	do {
		struct sk_buff *nskb = segs->next;
		if (err == 0) {
			err = __enqueue_packet_gso(segs, entry, queuenum);
		}

		if (err == 0) {
			queued++;
		}
		else {
			kfree_skb(segs);
		}

		segs = nskb;
	} while (segs);

	if (queued) {
		if (err) { /* some segments are already queued */
			free_entry(entry);
		}
		kfree_skb(skb);
		return 0;
	}

out_err:
	nf_bridge_adjust_segmented_data(skb);
	return err;
}

/* end of borrowing (& tweaking) this is borrowed by borrowes from net/netfilter/nfnetlink_queue_core.c */

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3,16,7))
#error "Check that signatures of this function and nf_register_queue_handler in file include/net/netfilter/nf_queue.h still match"
#endif
TRAMPOLINE(void nf_register_queue_handler_trampoline(const struct nf_queue_handler *qh))
TRAMPOLINE(void nf_unregister_queue_handler_trampoline(void))

void nf_register_queue_handler_detour(const struct nf_queue_handler *qh)
{
	mutex_lock(&queue_handler_mutex);

	WARN_ON(rcu_access_pointer(queue_handler[0]));
	if (queue_handler[0] != qh && ! queue_handler[0]) {
		rcu_assign_pointer(queue_handler[0], qh);
	}

	mutex_unlock(&queue_handler_mutex);

	TRACE(LOG_INFO, "%s: pid %d", __func__, current->pid);
}

void nf_unregister_queue_handler_detour(void)
{
	if (queue_handler[0]) {
		TRACE(LOG_INFO, "%s: unregistering hooked queue handler.", __func__);
		mutex_lock(&queue_handler_mutex);
		rcu_assign_pointer(queue_handler[0], NULL);
		mutex_unlock(&queue_handler_mutex);
		synchronize_rcu();
	}
}

void ip_filter_snort_stopped(pid_t pid)
{
	TRACE(LOG_INFO, "%s: Snort stopped. pid %d", __func__, pid);
	snort_ready = 0;
}

void ip_filter_snort_started(pid_t pid)
{
	TRACE(LOG_INFO, "%s: Snort started. pid %d", __func__, pid);
	snort_ready = 1;
}

static void nf_hook_drop(struct net *net, struct nf_hook_ops *hook)
{
	TRACE(LOG_INFO, "%s: Drop called, net: %p, hook: %p", __func__, net, hook);
}

static struct nf_queue_handler nfqh = {
	.outfn	= &enqueue_packet,
	.nf_hook_drop = &nf_hook_drop
};

#define IP4_ENABLED 	1
#define IP6_ENABLED 	1
#define SNORT_ENABLED 	1

static struct nf_hook_ops ipt_ops[] = {
#if IP4_ENABLED
	{
		.hook		= ipv4_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET,
		.hooknum	= NF_IP_PRE_ROUTING,
		.priority	= NF_IP_PRI_LAST,
	},
	{
		.hook		= ipv4_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET,
		.hooknum	= NF_IP_POST_ROUTING,
		.priority	= NF_IP_PRI_FIRST,
	},
#endif
#if IP6_ENABLED
	{
		.hook		= ipv6_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET6,
		.hooknum	= NF_IP_PRE_ROUTING,
		.priority	= NF_IP_PRI_LAST,
	},
	{
		.hook		= ipv6_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET6,
		.hooknum	= NF_IP_POST_ROUTING,
		.priority	= NF_IP_PRI_FIRST,
	},
#endif
#if SNORT_ENABLED
	{
		.hook		= ip_snort_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET,
		.hooknum	= NF_IP_PRE_ROUTING,
		.priority	= NF_IP_PRI_FIRST,
	},
	{
		.hook		= ip_snort_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET,
		.hooknum	= NF_IP_POST_ROUTING,
		.priority	= NF_IP_PRI_LAST,
	},
#if IP6_ENABLED
	{
		.hook		= ip_snort_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET6,
		.hooknum	= NF_IP_PRE_ROUTING,
		.priority	= NF_IP_PRI_FIRST,
	},
	{
		.hook		= ip_snort_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET6,
		.hooknum	= NF_IP_POST_ROUTING,
		.priority	= NF_IP_PRI_LAST,
	},
#endif
#endif
};

int ip_filter_init(void)
{
	int err = 0;

	rcu_assign_pointer(queue_handler[0], NULL);
	queue_init(&g_main_queue);
	queue_init(&g_pending_queue);

	nf_register_queue_handler(&nfqh);

	TRACE(LOG_INFO, "%s: enqueue packet handler registered.", __func__);
	nf_queue_handler_registered = 1;

	/* register netfilter hooks */
	if ((err = nf_register_hooks(ipt_ops, ARRAY_SIZE(ipt_ops)))) {
		TRACE(LOG_ERR, "%s: nf_register_hooks(): %d", __func__, err);
		return err;
	}
	TRACE(LOG_INFO, "%s: netfilter hooks registered.", __func__);
	nf_hooks_registered = 1;

	/* hooking nf_register_queue_handler() */
	TRACE(LOG_INFO, "%s: hooking nf_register_queue_handler() function...", __func__);
	if ((nf_hook1_bytes_to_replace = hook_install(nf_register_queue_handler, nf_register_queue_handler_detour, nf_register_queue_handler_trampoline)) < 0) {
		return nf_hook1_bytes_to_replace;
	}

	/* hooking nf_unregister_queue_handler() */
	TRACE(LOG_INFO, "%s: hooking nf_unregister_queue_handler() function...", __func__);
	if ((nf_hook2_bytes_to_replace = hook_install(nf_unregister_queue_handler, nf_unregister_queue_handler_detour, nf_unregister_queue_handler_trampoline)) < 0) {
		hook_remove(nf_register_queue_handler, nf_register_queue_handler_trampoline, nf_hook1_bytes_to_replace);
		return nf_hook2_bytes_to_replace;
	}

	rcu_assign_pointer(nf_nat_decode_session_hook, nat_decode_session);

	TRACE(LOG_INFO, "%s: done.", __func__);

	return 0;
}

int ip_filter_close(void)
{
	TRACE(LOG_INFO, "%s: unhooking nf_register_queue_handler() function...", __func__);
	if ( ! hook_remove(nf_register_queue_handler, nf_register_queue_handler_trampoline, nf_hook1_bytes_to_replace)) {
		TRACE(LOG_INFO, "%s: done.", __func__);
	}
	TRACE(LOG_INFO, "%s: unhooking nf_unregister_queue_handler() function...", __func__);
	if ( ! hook_remove(nf_unregister_queue_handler, nf_unregister_queue_handler_trampoline, nf_hook2_bytes_to_replace)) {
		TRACE(LOG_INFO, "%s: done.", __func__);
	}


	if (nf_hooks_registered) {
		nf_unregister_hooks(ipt_ops, ARRAY_SIZE(ipt_ops));
		TRACE(LOG_INFO, "%s: netfilter hooks unregistered.", __func__);
	}
	if (nf_queue_handler_registered) {
		nf_unregister_queue_handler();
		TRACE(LOG_INFO, "%s: enqueue packet handler unregistered.", __func__);
	}

	queue_close(&g_pending_queue);
	queue_close(&g_main_queue);

	nf_hooks_registered = 0;
	nf_queue_handler_registered = 0;
	nf_hook1_bytes_to_replace = 0;
	nf_hook2_bytes_to_replace = 0;

	return 0;
}
