/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <linux/if_vlan.h>

#include "hdrs.h"
#include "hooks.h"

#define NETDEV_EXTENSION_TAG	0x12345678

static int netdev_event(struct notifier_block *this, unsigned long event, void *ptr);

/**
 *	Locals
 */
static int gs_netdev_notifier_registered = 0;
static int gs_inet_notifier_registered = 0;
static int gs_inet6_notifier_registered = 0;

static struct notifier_block gs_ipf_netdev_notifier = {
	.notifier_call = netdev_event,
};
static struct notifier_block gs_ipf_inet_notifier = {
	.notifier_call	= netdev_event,
};
static struct notifier_block gs_ipf_inet6_notifier = {
	.notifier_call	= netdev_event,
};

/* [Bug 58820] -  GSO enabled on bridge => packets dropped in kipf => slow internet */
typedef int (*bridge_ioctl_handler)(struct net_device *dev, struct ifreq *rq, int cmd);
bridge_ioctl_handler old_bridge_ioctl_handler = NULL;

static struct net_device_extension*
netdev_alloc_extension(
	struct net_device *dev
	)
{
	struct net_device_extension *ext = NULL;

	/* re-use of Econet protocol pointer to specific data */
	if (dev->ax25_ptr) {

		ext = dev->ax25_ptr;

		if (ext->tag != NETDEV_EXTENSION_TAG) {
			TRACE(LOG_ERR, "%s: pointer to Econet protocol specific data on '%s' can't be re-used.", __func__, dev->name);
			return NULL;
		}

		TRACE(LOG_INFO, "%s: device extension %p on '%s' already allocated.", __func__, ext, dev->name);
		return ext;

	} else {
		if (!(ext = kmalloc(sizeof(struct net_device_extension), GFP_ATOMIC))) {
			TRACE(LOG_ERR, "%s: can't allocate device extension on '%s'.", __func__, dev->name);
			return NULL;
		}
	}

	ext->tag = NETDEV_EXTENSION_TAG;
	ext->mode = NF_IFACE_MODE_DISABLED;

	/* save */
	dev->ax25_ptr = ext;

	TRACE(LOG_INFO, "%s: device extension %p on '%s' allocated.", __func__, ext, dev->name);

	return ext;
}

static void
netdev_free_extension(
	struct net_device *dev
	)
{
	if (dev && dev->ax25_ptr && ((struct net_device_extension*)dev->ax25_ptr)->tag == NETDEV_EXTENSION_TAG) {
		TRACE(LOG_INFO, "%s: device extension %p on '%s' freed.", __func__, dev->ax25_ptr, dev->name);
		kfree(dev->ax25_ptr);
		dev->ax25_ptr = NULL;
	}
}

struct net_device_extension*
netdev_get_extension(
	const struct net_device *dev
	)
{
	if (dev && dev->ax25_ptr && ((struct net_device_extension*)dev->ax25_ptr)->tag == NETDEV_EXTENSION_TAG) {
		return dev->ax25_ptr;
	}
	return NULL;
}

/**
 *	[Bug 58820] -  GSO enabled on bridge => packets dropped in kipf => slow internet
 */
static void
netdev_fix_bridge_ioctl_handler(
	void
	)
{
	struct net_device *dev = NULL;
	/* nothing to fix */
	if (!old_bridge_ioctl_handler) {
		return;
	}
	read_lock(&dev_base_lock);
	for_each_netdev(&init_net, dev) {
		if (dev->priv_flags & IFF_EBRIDGE) {
			bridge_ioctl_handler *p;
			if (!dev->netdev_ops || !dev->netdev_ops->ndo_do_ioctl) {
				TRACE(LOG_WARN, "%s: %s hasn't ioctl handler?!?", __func__, dev->name);
				continue;
			}
			p = (bridge_ioctl_handler*)&dev->netdev_ops->ndo_do_ioctl;
			*p = old_bridge_ioctl_handler;
			TRACE(LOG_INFO, "%s: ioctl handler on %s fixed.", __func__, dev->name);
		}
	}
	read_unlock(&dev_base_lock);
	old_bridge_ioctl_handler = NULL;
}

static void
netdev_remove_all_extensions(
	void
	)
{
	struct net_device *dev = NULL;

	read_lock(&dev_base_lock);

	for_each_netdev(&init_net, dev) {
		netdev_free_extension(dev);
	}

	read_unlock(&dev_base_lock);
}

// TODO: use dev_getbyhwaddr_rcu()?
struct net_device *
netdev_find_by_mac(
	unsigned char *mac,
	int len
	)
{
	struct net_device *dev = NULL;

	read_lock(&dev_base_lock);

	for_each_netdev(&init_net, dev) {
		if (!memcmp(dev->dev_addr, mac, MIN(dev->addr_len, len))) {
			read_unlock(&dev_base_lock);
			return dev;
		}
	}

	read_unlock(&dev_base_lock);

	return NULL;
}

// TODO: use dev_get_by_index()?
struct net_device *
netdev_find_by_ifindex(
	unsigned int ifindex
	)
{
	struct net_device *dev = NULL;

	read_lock(&dev_base_lock);

	for_each_netdev(&init_net, dev) {
		if (dev->ifindex == ifindex) {
			read_unlock(&dev_base_lock);
			return dev;
		}
	}

	read_unlock(&dev_base_lock);

	return NULL;
}

void
netdev_enum_all(
	void
	)
{
	struct net_device *dev = NULL;

	read_lock(&dev_base_lock);

	for_each_netdev(&init_net, dev) {
		TRACE(LOG_INFO,
			"%s: '%s': dev=%p, mac=%02x:%02x:%02x:%02x:%02x:%02x, mtu=%d, state=%d, features=0x%X, ifindex=0x%X, "
			"flags=0x%X, priv_flags=0x%X, type=%d, hard_header_len=%d, addr_len=%d",
			__func__,
			dev->name,
			dev,
			dev->dev_addr[0],
			dev->dev_addr[1],
			dev->dev_addr[2],
			dev->dev_addr[3],
			dev->dev_addr[4],
			dev->dev_addr[5],
			dev->mtu,
			dev->state,
			dev->features,
			dev->ifindex,
			dev->flags,
			dev->priv_flags,
			dev->type,
			dev->hard_header_len,
			dev->addr_len
			);
	}

	read_unlock(&dev_base_lock);
}

static uint getHwType(unsigned short dev_type)
{
	switch (dev_type) {
		case ARPHRD_ETHER:
		case ARPHRD_SIT:
			return KIPF_HW_TYPE_ETH;
		case ARPHRD_PPP:
			return KIPF_HW_TYPE_PPP;
	};
	return KIPF_HW_TYPE_UNK;
}

static bool emptyHwAddr(char *hwAddr)
{
	static char emptyAddr[6] = {0};
	return memcmp(emptyAddr, hwAddr, sizeof(emptyAddr)) == 0;
}

static struct sk_buff*
netdev_prepare(
	struct net_device *dev
	)
{
	struct sk_buff *skb;
	struct kipf_iface *iface;

	/* We don't want to see loopback ifaces */
	if (dev->flags & IFF_LOOPBACK) {
		TRACE(LOG_WARN, "%s: '%s' is loopback device.", __func__, dev->name);
		return NULL;
	}

	/* We don't want to see ifaces with another type then ethernet or ppp */
	if (KIPF_HW_TYPE_UNK == getHwType(dev->type)) {
		TRACE(LOG_WARN, "%s: '%s' isn't ethernet or ppp device. (%d)", __func__, dev->name, dev->type);
		return NULL;
	}

	// Store data from dev into skb.
	skb = alloc_skb(sizeof(struct kipf_iface), GFP_ATOMIC);

	if (!skb) {
		TRACE(LOG_WARN, "%s: dev_alloc_skb(): can't allocate memory for dev='%s'.", __func__, dev->name);
		return NULL;
	}

	init_kipf_skb_data(skb, NET_DEV_PACKET);
	iface = (struct kipf_iface *) skb->data;

	iface->header.magic = KIPF_MAGIC_IFACE;
	iface->mtu = dev->mtu;
	iface->hw_type = getHwType(dev->type);
	iface->ifindex = dev->ifindex;

	iface->flags = 0;
	if (dev->flags & IFF_UP) {
		iface->flags |= KIPF_IFACE_UP;
	}
	if (netif_carrier_ok(dev)) {
		iface->flags |= KIPF_CARRIER_UP;
	}

	if (dev->type != ARPHRD_SIT && dev->type != ARPHRD_PPP) {
		memcpy(iface->hw_addr, dev->dev_addr, MIN(sizeof(iface->hw_addr), dev->addr_len));
	}
	else {
		memset(iface->hw_addr, 0, sizeof(iface->hw_addr));
	}

	memcpy(iface->name, dev->name, MIN(sizeof(iface->name), strlen(dev->name) + 1));

	/* create id */
	if (dev->type == ARPHRD_PPP) {
		name2IfaceId(IIT_Ppp, iface->id, sizeof(iface->id), dev->name);
	} else if (dev->type == ARPHRD_SIT) {
		name2IfaceId(IIT_Sit, iface->id, sizeof(iface->id), dev->name);
	} else if (dev->priv_flags & IFF_EBRIDGE) {
		name2IfaceId(IIT_Bridge, iface->id, sizeof(iface->id), dev->name);
	} else if (is_vlan_dev(dev)) {
		struct net_device *parent = vlan_dev_real_dev(dev);
		if (unlikely( ! parent)) {
			mac2IfaceId(IIT_Vlan, iface->id, sizeof(iface->id), emptyHwAddr(dev->perm_addr) ? dev->dev_addr : dev->perm_addr, vlan_dev_vlan_id(dev), NULL);
		}
		else {
			mac2IfaceId(IIT_Vlan, iface->id, sizeof(iface->id), emptyHwAddr(parent->perm_addr) ? parent->dev_addr : parent->perm_addr, vlan_dev_vlan_id(dev), NULL);
		}
	} else if (dev->ieee80211_ptr) {
		mac2IfaceId(IIT_Wlan, iface->id, sizeof(iface->id), emptyHwAddr(dev->perm_addr) ? dev->dev_addr : dev->perm_addr, 0, dev->name);
	} else {
		mac2IfaceId(IIT_Eth, iface->id, sizeof(iface->id), emptyHwAddr(dev->perm_addr) ? dev->dev_addr : dev->perm_addr, 0, NULL);
	}

	return skb;
}

static void
netdev_handle_event(
	struct net_device *dev,
	unsigned long event
	)
{
	struct sk_buff *skb = NULL;
	int err = 0;

	if (event == NETDEV_UP) {
		if (!netdev_alloc_extension(dev)) {
			return;
		}
	// commented-out - Bug 52002 - Any interface traffic is dropped after secondary IP address is removed
	// } else {
	//	netdev_free_extension(dev);
	//
	}

	if (!(skb = netdev_prepare(dev))) {
		return;
	}

	if ((err = queue_insert(&g_main_queue, skb, 1))) {

		/* It's normal that queue_insert() will fail after module load, because queue is not enabled. */
		TRACE(LOG_WARN, "netdev_handle_event: queue_insert(): dev='%s', err=%d", dev->name, err);

		kfree_skb(skb);
		return;
	}

	io_dev_wakeup();
}

void
netdev_signal_end(
	void
	)
{
	struct sk_buff *skb = NULL;
	struct kipf_iface *iface;
	int err = 0;

	skb = alloc_skb(sizeof(struct kipf_iface), GFP_ATOMIC);
	if (!skb) {
		TRACE(LOG_WARN, "%s: alloc_skb(): can't allocate memory.", __func__);
		return;
	}

	init_kipf_skb_data(skb, NET_DEV_PACKET);
	iface = (struct kipf_iface *) skb->data;
	iface->header.magic = KIPF_MAGIC_IFACE;
	iface->ifindex = 0;

	if ((err = queue_insert(&g_main_queue, skb, 1))) {
		TRACE(LOG_WARN, "%s: queue_insert(): err=%d", __func__, err);
		kfree_skb(skb);
		return;
	}

	TRACE(LOG_INFO, "%s: null net device signaled.", __func__);
	io_dev_wakeup();
}

static void
netdev_dump_event(
	struct net_device *dev,
	int event
	)
{
	TRACE(LOG_INFO, "netdev_event(): %d: '%s', if_index %d, refcnt %d, dev %p, mac %02x:%02x:%02x:%02x:%02x:%02x, mtu %d%s",
		event,
		dev->name,
		dev->ifindex,
		netdev_refcnt_read(dev),
		dev,
		dev->dev_addr[0],
		dev->dev_addr[1],
		dev->dev_addr[2],
		dev->dev_addr[3],
		dev->dev_addr[4],
		dev->dev_addr[5],
		dev->mtu,
		((dev->priv_flags & IFF_EBRIDGE) ? (", bridge") : (""))
		);
}

void
netdev_signal_all_up(
	void
	)
{
	struct net_device *dev = NULL;

	TRACE(LOG_INFO, "%s: signaling all IFF_UP flaged net devices...", __func__);

	read_lock(&dev_base_lock);

	for_each_netdev(&init_net, dev) {

		/* We dont want to see down ifaces */
		if (!(dev->flags & IFF_UP)) {
			continue;
		}

		netdev_dump_event(dev, NETDEV_UP);
		netdev_handle_event(dev, NETDEV_UP);
	}

	netdev_signal_end();

	read_unlock(&dev_base_lock);
}

/**
 * [Bug 58820] - GSO enabled on bridge => packets dropped in kipf => slow internet
 * we need NETDEV_UP event so user space disable GSO on bridge again
 */
int br_dev_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	int ret;
	if (!old_bridge_ioctl_handler) {
		TRACE(LOG_ERR, "%s: original bridge ioctl handler on %s is null?!?", __func__, dev->name);
		return -EINVAL;
	}
	ret = old_bridge_ioctl_handler(dev, rq, cmd);
	if (!ret && (cmd == SIOCBRADDIF || cmd == SIOCBRDELIF)) {
		TRACE(LOG_INFO, "%s: interface %s %s", __func__, cmd == SIOCBRADDIF ? "added to" : "removed from", dev->name);
		netdev_handle_event(dev, NETDEV_UP);
	} 
	return ret;
}

static int netdev_event(struct notifier_block *this, unsigned long event, void *ptr)
{
	struct sk_buff *skb = NULL;
	struct net_device *dev = NULL;
	if (this == &gs_ipf_inet_notifier) {
		dev = ((struct in_ifaddr *)ptr)->ifa_dev->dev;
	} else if (this == &gs_ipf_inet6_notifier) {
		dev = ((struct inet6_ifaddr *)ptr)->idev->dev;
	} else {
		dev = ((struct netdev_notifier_info *)ptr)->dev;
	}

	switch (event) {
		case NETDEV_UP:
		case NETDEV_DOWN:
		case NETDEV_CHANGE:
			netdev_dump_event(dev, event);
			netdev_handle_event(dev, event);
			break;

		case NETDEV_CHANGEADDR: // ?!
		case NETDEV_GOING_DOWN: // ?
		case NETDEV_REBOOT:
		case NETDEV_CHANGEMTU:
		case NETDEV_CHANGENAME:
		case NETDEV_FEAT_CHANGE:
			netdev_dump_event(dev, event);
			break;

		case NETDEV_REGISTER:
			netdev_dump_event(dev, event);
			/*
   			 * [Bug 58820] -  GSO enabled on bridge => packets dropped in kipf => slow internet
 			 * - adding or removing interface from bridge caused offloading to be reenabed on bridge
			 * - we hooked bridge ioctl handler and after successful remove/add ioctl we will notify
			 *   wr about "interface up" event - it will leads to disabling all offloading on device
  			 */
			/* bridge is up ? */
			if (dev->priv_flags & IFF_EBRIDGE) {
				bridge_ioctl_handler *p;
				TRACE(LOG_INFO, "%s: hooking %s ioctl handler.", __func__, dev->name);
				if (!dev->netdev_ops || !dev->netdev_ops->ndo_do_ioctl) {
					TRACE(LOG_ERR, "%s: %s hasn't ioctl handler?!?", __func__, dev->name);
					break;
				}
				/* save old ioctl handler */
				if (!old_bridge_ioctl_handler) {
					old_bridge_ioctl_handler = dev->netdev_ops->ndo_do_ioctl;
				}
				p = (bridge_ioctl_handler*)&dev->netdev_ops->ndo_do_ioctl;
				if (hook_prepare_func(p, sizeof(void*))) {
					break;
				}
				*p = br_dev_ioctl;
				TRACE(LOG_INFO, "%s: ioctl handler on %s hooked.", __func__, dev->name);
			}
			break;
		case NETDEV_UNREGISTER:
			netdev_dump_event(dev, event);
			if (this != &gs_ipf_netdev_notifier) {
				break;
			}

			while ((skb = queue_remove(&g_main_queue, dev, queue_cmp_by_dev))) {
				TRACE(LOG_INFO, "%s: skb %p removed from main_queue.", __func__, skb);
				pkt_verdict(skb, NF_DROP);
			}

			while ((skb = queue_remove(&g_pending_queue, dev, queue_cmp_by_dev))) {
				TRACE(LOG_INFO, "%s: skb %p removed from pending_queue.", __func__, skb);
				pkt_verdict(skb, NF_DROP);
			}

			break;

		default:
			TRACE(LOG_WARN, "%s: net device event 0x%04X on dev '%s' signaled.", __func__, event, dev->name);
			break;
	}
	return NOTIFY_DONE;
}

int netdev_close(void)
{
	int err = 0;
	if (gs_netdev_notifier_registered) {
		if ((err = unregister_netdevice_notifier(&gs_ipf_netdev_notifier))) {
			TRACE(LOG_WARN, "%s: unregister_netdevice_notifier(): %d", __func__, err);
		} else {
			TRACE(LOG_INFO, "%s: net device notifier unregistered.", __func__);
		}
		gs_netdev_notifier_registered = 0;
	}
	if (gs_inet_notifier_registered) {
		if ((err = unregister_inetaddr_notifier(&gs_ipf_inet_notifier))) {
			TRACE(LOG_WARN, "%s: unregister_inetaddr_notifier(): %d", __func__, err);
		} else {
			TRACE(LOG_INFO, "%s: inet addr notifier unregistered.", __func__);
		}
		gs_inet_notifier_registered = 0;
	}
	if (gs_inet6_notifier_registered) {
		if ((err = unregister_inet6addr_notifier(&gs_ipf_inet6_notifier))) {
			TRACE(LOG_WARN, "%s: unregister_inet6addr_notifier(): %d", __func__, err);
		} else {
			TRACE(LOG_INFO, "%s: inet6 addr notifier unregistered.", __func__);
		}
		gs_inet6_notifier_registered = 0;
	}
	netdev_remove_all_extensions();
	/* [Bug 58820] -  GSO enabled on bridge => packets dropped in kipf => slow internet */
	netdev_fix_bridge_ioctl_handler();
	return 0;
}

int netdev_init(void)
{
	int err = 0;
	/* register net devices notifier */
	if ((err = register_netdevice_notifier(&gs_ipf_netdev_notifier))) {
		TRACE(LOG_ERR, "%s: register_netdevice_notifier(): %d", __func__, err);
		return err;
	}
	TRACE(LOG_INFO, "%s: net device notifier registered.", __func__);
	gs_netdev_notifier_registered = 1;

	/* register IPv4 address change notifier */
	if ((err = register_inetaddr_notifier(&gs_ipf_inet_notifier))) {
		TRACE(LOG_ERR, "%s: register_inetaddr_notifier(): %d", __func__, err);
		return err;
	}
	TRACE(LOG_INFO, "%s: inet addr notifier registered.", __func__);
	gs_inet_notifier_registered = 1;

	/* register IPv6 address change notifier */
	if ((err = register_inet6addr_notifier(&gs_ipf_inet6_notifier))) {
		TRACE(LOG_ERR, "%s: register_inet6addr_notifier(): %d", __func__, err);
		return err;
	}
	TRACE(LOG_INFO, "%s: inet6 addr notifier registered.", __func__);
	gs_inet6_notifier_registered = 1;

	return err;
}
